/*
 * Copyright (c) 2017 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <router.h>
#include <props.h>

#define MAX_NPROPS 2

typedef struct _plugstate_t plugstate_t;
typedef struct _plughandle_t plughandle_t;

struct _plugstate_t {
	int32_t gate_toggle;
	int32_t gate_state;
};

struct _plughandle_t {
	LV2_URID_Map *map;

	const LV2_Atom_Sequence *event_in;
	craft_t event_out;
	const LV2_Atom_Sequence *control;
	craft_t notify;

	struct {
		LV2_URID deflector_gateToggle;
		LV2_URID deflector_gateState;
	} urid;

	plugstate_t state;
	plugstate_t stash;

	PROPS_T(props, MAX_NPROPS);
};

static void
_intercept_gate_toggle(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.gate_toggle)
	{
		handle->state.gate_toggle = false;
		handle->state.gate_state = !handle->state.gate_state;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.deflector_gateToggle, &handle->notify.ref);
		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.deflector_gateState, &handle->notify.ref);
	}
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = ROUTER__gateToggle,
		.offset = offsetof(plugstate_t, gate_toggle),
		.type = LV2_ATOM__Bool,
		.event_cb = _intercept_gate_toggle
	},
	{
		.property = ROUTER__gateState,
		.offset = offsetof(plugstate_t, gate_state),
		.type = LV2_ATOM__Bool
	}
};

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate __attribute__((unused)),
	const char *bundle_path __attribute__((unused)),
	const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;
	mlock(handle, sizeof(plughandle_t));

	for(unsigned i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
	}

	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	_craft_init(&handle->notify, handle->map);
	_craft_init(&handle->event_out, handle->map);

	handle->urid.deflector_gateToggle = handle->map->map(handle->map->handle,
		ROUTER__gateToggle);
	handle->urid.deflector_gateState = handle->map->map(handle->map->handle,
		ROUTER__gateState);

	if(!props_init(&handle->props, descriptor->URI,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "failed to initialize property structure\n");
		free(handle);
		return NULL;
	}

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = (plughandle_t *)instance;

	switch(port)
	{
		case 0:
			_craft_connect(&handle->event_out, data);
			break;
		case 1:
			handle->event_in = (const void *)data;
			break;

		case 2:
			handle->control = (const void *)data;
			break;
		case 3:
			_craft_connect(&handle->notify, data);
			break;
		default:
			break;
	}
}

static void
run(LV2_Handle instance, uint32_t nsamples __attribute__((unused)))
{
	plughandle_t *handle = instance;

	_craft_in(&handle->notify);
	_craft_in(&handle->event_out);

	props_idle(&handle->props, &handle->notify.forge, 0, &handle->notify.ref);

	int64_t from = 0;

	LV2_ATOM_SEQUENCE_FOREACH(handle->control, evc)
	{
		craft_t *craft = &handle->event_out;
		const int64_t to = evc->time.frames;

		if(handle->state.gate_state)
		{
			LV2_ATOM_SEQUENCE_FOREACH(handle->event_in, evi) //FIXME inefficient
			{
				const LV2_Atom *atom = &evi->body;

				if(evi->time.frames < from)
					continue; // too late, skip step

				if(evi->time.frames >= to)
					break; // too early to schedule, break loop

				if(craft->ref)
					craft->ref = lv2_atom_forge_frame_time(&craft->forge, evi->time.frames);
				if(craft->ref)
					craft->ref = lv2_atom_forge_write(&craft->forge, atom, lv2_atom_total_size(atom));
			}
		}

		const LV2_Atom *atom = &evc->body;
		const LV2_Atom_Object *obj = (const LV2_Atom_Object *)atom;

		props_advance(&handle->props, &handle->notify.forge, to, obj, &handle->notify.ref);
		from = to;
	}

	if(handle->state.gate_state)
	{
		craft_t *craft = &handle->event_out;

		LV2_ATOM_SEQUENCE_FOREACH(handle->event_in, evi) //FIXME inefficient
		{
			const LV2_Atom *atom = &evi->body;

			if(evi->time.frames < from)
				continue; // too late, skip step

			if(craft->ref)
				craft->ref = lv2_atom_forge_frame_time(&craft->forge, evi->time.frames);
			if(craft->ref)
				craft->ref = lv2_atom_forge_write(&craft->forge, atom, lv2_atom_total_size(atom));
		}
	}

	_craft_out(&handle->notify);
	_craft_out(&handle->event_out);
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	munlock(handle, sizeof(plughandle_t));
	free(handle);
}

static LV2_State_Status
_state_save(LV2_Handle instance, LV2_State_Store_Function store,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_save(&handle->props, store, state, flags, features);
}

static LV2_State_Status
_state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_restore(&handle->props, retrieve, state, flags, features);
}

static const LV2_State_Interface state_iface = {
	.save = _state_save,
	.restore = _state_restore
};

static const void*
extension_data(const char* uri)
{
	if(!strcmp(uri, LV2_STATE__interface))
		return &state_iface;

	return NULL;
}

const LV2_Descriptor router_atom_deflector = {
	.URI						= ROUTER__ATOM_DEFLECTOR,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};
