/*
 * Copyright (c) 2017 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <router.h>
#include <props.h>

#define MAX_NOUTS 8
#define MAX_NPROPS 5

typedef struct _plugstate_t plugstate_t;
typedef struct _plughandle_t plughandle_t;

struct _plugstate_t {
	int32_t next_index;
	int32_t prev_index;
	int32_t roll_under_index;
	int32_t roll_over_index;
	int32_t current_index;
};

struct _plughandle_t {
	LV2_URID_Map *map;

	const LV2_Atom_Sequence *event_in;
	craft_t event_out [MAX_NOUTS];
	const LV2_Atom_Sequence *control;
	craft_t notify;

	struct {
		LV2_URID sprinkler_nextIndex;
		LV2_URID sprinkler_prevIndex;
		LV2_URID sprinkler_currentIndex;
	} urid;

	plugstate_t state;
	plugstate_t stash;

	PROPS_T(props, MAX_NPROPS);
};

static void
_intercept_next_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.next_index)
	{
		handle->state.next_index = false;

		handle->state.current_index += 1;
		if(handle->state.current_index > handle->state.roll_over_index)
			handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_nextIndex, &handle->notify.ref);
		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_prev_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.prev_index)
	{
		handle->state.prev_index = false;

		handle->state.current_index -= 1;
		if(handle->state.current_index < handle->state.roll_under_index)
			handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_prevIndex, &handle->notify.ref);
		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_roll_under_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index < handle->state.roll_under_index)
	{
		handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_roll_over_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index > handle->state.roll_over_index)
	{
		handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_current_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index > handle->state.roll_over_index)
	{
		handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
	else if(handle->state.current_index < handle->state.roll_under_index)
	{
		handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.sprinkler_currentIndex, &handle->notify.ref);
	}
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = ROUTER__nextIndex,
		.offset = offsetof(plugstate_t, next_index),
		.type = LV2_ATOM__Bool,
		.event_cb = _intercept_next_index
	},
	{
		.property = ROUTER__prevIndex,
		.offset = offsetof(plugstate_t, prev_index),
		.type = LV2_ATOM__Bool,
		.event_cb = _intercept_prev_index
	},
	{
		.property = ROUTER__rollUnderIndex,
		.offset = offsetof(plugstate_t, roll_under_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_roll_under_index
	},
	{
		.property = ROUTER__rollOverIndex,
		.offset = offsetof(plugstate_t, roll_over_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_roll_over_index
	},
	{
		.property = ROUTER__currentIndex,
		.offset = offsetof(plugstate_t, current_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_current_index
	}
};

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate __attribute__((unused)),
	const char *bundle_path __attribute__((unused)),
	const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;
	mlock(handle, sizeof(plughandle_t));

	for(unsigned i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
	}

	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	_craft_init(&handle->notify, handle->map);
	for(unsigned i = 0; i < MAX_NOUTS; i++)
		_craft_init(&handle->event_out[i], handle->map);

	handle->urid.sprinkler_nextIndex = handle->map->map(handle->map->handle,
		ROUTER__nextIndex);
	handle->urid.sprinkler_prevIndex = handle->map->map(handle->map->handle,
		ROUTER__prevIndex);
	handle->urid.sprinkler_currentIndex = handle->map->map(handle->map->handle,
		ROUTER__currentIndex);

	if(!props_init(&handle->props, descriptor->URI,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "failed to initialize property structure\n");
		free(handle);
		return NULL;
	}

	handle->state.current_index = 1;
	handle->state.roll_over_index = 1;

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = (plughandle_t *)instance;

	switch(port)
	{
		case 0:
			handle->event_in = (const void *)data;
			break;
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
			_craft_connect(&handle->event_out[port-1], data);
			break;
		case 9:
			handle->control = (const void *)data;
			break;
		case 10:
			_craft_connect(&handle->notify, data);
			break;
		default:
			break;
	}
}

static void
run(LV2_Handle instance, uint32_t nsamples __attribute__((unused)))
{
	plughandle_t *handle = instance;

	_craft_in(&handle->notify);
	for(unsigned i = 0; i < MAX_NOUTS; i++)
		_craft_in(&handle->event_out[i]);

	props_idle(&handle->props, &handle->notify.forge, 0, &handle->notify.ref);

	const LV2_Atom_Event *evi = lv2_atom_sequence_begin(&handle->event_in->body);
	LV2_ATOM_SEQUENCE_FOREACH(handle->control, evc)
	{
		craft_t *craft = &handle->event_out[handle->state.current_index - 1];
		const int64_t to = evc->time.frames;

		while(!lv2_atom_sequence_is_end(&handle->event_in->body, handle->event_in->atom.size, evi))
		{
			const LV2_Atom *atom = &evi->body;

			if(evi->time.frames >= to)
				break; // too early to schedule, break loop

			if(craft->ref)
				craft->ref = lv2_atom_forge_frame_time(&craft->forge, evi->time.frames);
			if(craft->ref)
				craft->ref = lv2_atom_forge_write(&craft->forge, atom, lv2_atom_total_size(atom));

			evi = lv2_atom_sequence_next(evi);
		}

		const LV2_Atom *atom = &evc->body;
		const LV2_Atom_Object *obj = (const LV2_Atom_Object *)atom;

		props_advance(&handle->props, &handle->notify.forge, to, obj, &handle->notify.ref);
	}

	{
		craft_t *craft = &handle->event_out[handle->state.current_index - 1];

		while(!lv2_atom_sequence_is_end(&handle->event_in->body, handle->event_in->atom.size, evi))
		{
			const LV2_Atom *atom = &evi->body;

			if(craft->ref)
				craft->ref = lv2_atom_forge_frame_time(&craft->forge, evi->time.frames);
			if(craft->ref)
				craft->ref = lv2_atom_forge_write(&craft->forge, atom, lv2_atom_total_size(atom));

			evi = lv2_atom_sequence_next(evi);
		}
	}

	_craft_out(&handle->notify);
	for(unsigned i = 0; i < MAX_NOUTS; i++)
		_craft_out(&handle->event_out[i]);
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	munlock(handle, sizeof(plughandle_t));
	free(handle);
}

static LV2_State_Status
_state_save(LV2_Handle instance, LV2_State_Store_Function store,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_save(&handle->props, store, state, flags, features);
}

static LV2_State_Status
_state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_restore(&handle->props, retrieve, state, flags, features);
}

static const LV2_State_Interface state_iface = {
	.save = _state_save,
	.restore = _state_restore
};

static const void*
extension_data(const char* uri)
{
	if(!strcmp(uri, LV2_STATE__interface))
		return &state_iface;

	return NULL;
}

const LV2_Descriptor router_atom_sprinkler = {
	.URI						= ROUTER__ATOM_SPRINKLER,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};
